from typing import List, Optional

from karp5.config import conf_mgr
from karp5.cli import upload_offline


def recover(mode: str, *, suffix: Optional[str] = None, lexicons: Optional[List[str]] = None):
    if not lexicons:
        lexicons = conf_mgr.get_lexiconlist(mode)

    index_name = upload_offline.make_indexname(mode, suffix)

    return "Ok"



"""[summary]
"""
from karp5.errors import KarpException


class KarpConfigException(KarpException):
    """[summary]
    """

    def __init__(self, msg_fmt, *msg_args,  debug_msg=None):
        if not msg_args:
            msg = msg_fmt
        else:
            msg = msg_fmt % msg_args
        super().__init__(msg, debug_msg=debug_msg)



""" Creates json annotated with metadata (index, type)
    that can be sent to elasticsearch
    Input: json structures that might be of type string ('{"hej" : "hu"}')
"""
import json
from typing import Dict, Iterable

from karp5.document import doc_to_es
from .errors import BulkifyError

# from karp5.document import doc_to_sql


_index = "test"
_type = "test"


def bulkify(data, bulk_info={}, with_id=False):
    """ parse a string with either a list or a single json object
        annotate with meta data
    """
    index = bulk_info.get("index", _index)
    itype = bulk_info.get("type", _type)
    items = json.loads(data)
    result = []
    for item in items:
        data_doc = item["_source"] if with_id else item
        # assert 'lexiconName' in data_doc, "document doesn't have the field 'lexiconName'"
        es_doc = doc_to_es(data_doc, data_doc["lexiconName"], "bulk")
        doc = {
            "_index": index,
            "_type": itype,
            "_source": es_doc,
        }
        if with_id:
            doc["_id"] = item["_id"]
        result.append(doc)
    return result


def bulkify_from_sql(entries: Iterable[Dict], index: str, index_type: str):
    """Format a iterable of entries mapped to sql objects into a bulk insert.

    Arguments:
        entries {Iterable[Dict]} -- entries to prepare for ES
        index {str} -- the index to add entries to
        index_type {str} -- the type of the index
    """
    """  """
    return (
        {
            "_index": index,
            "_type": index_type,
            "_id": entry["id"],
            "_source": doc_to_es(entry["doc"], entry["doc"]["lexiconName"], "bulk"),
        }
        for entry in entries
    )


def bulkify_sql(data, bulk_info={}):
    """ format a dictionary of ids mapped to sql objects into a bulk insert """
    index = bulk_info.get("index", _index)
    itype = bulk_info.get("type", _type)
    return [
        {"_index": index, "_type": itype, "_id": _id, "_source": item["doc"]}
        for _id, item in data.items()
        if item["status"] != "removed"
    ]
    # result = []
    # for _id, item in data.items():
    #     if item["status"] != "removed":
    #         data_doc = item["doc"]
    #         es_doc = doc_to_es(data_doc, data_doc["lexiconName"], "bulk")
    #         doc = {"_index": index, "_type": itype, "_source": es_doc, "_id": _id}
    #         result.append(doc)
    # return result



import collections
import copy
import json

from karp5.server.translator.validatejson import escape, validate_json


def test_validate_json_on_panacea(app):
    file_name = "karp5/tests/data/data/panacea/panacea.json"
    lexicon_id = "panacea"

    with open(file_name) as fp:
        objs = json.load(fp)
        for obj in objs:
            orig = copy.deepcopy(obj)
            validate_json(obj, lexicon_id)
            compare(orig, obj)


def compare(orig, obj):
    assert type(orig) == type(obj)

    if isinstance(orig, str):
        # check that strings are escaped but otherwise unchanged
        assert escape(orig) == obj
        return

    if isinstance(orig, collections.Sized):
        # all other types of objects should have unchanged length
        assert len(orig) == len(obj)

    # recursively check the rest
    if isinstance(orig, dict):
        for key, val in orig.items():
            compare(val, obj[key])
    elif isinstance(orig, list):
        for i, val in enumerate(orig):
            compare(val, obj[i])
    else:
        assert orig == obj



# from karp5 import create_app

import karp5
from karp5.config import conf_mgr

# def test_config():
#     assert not create_app().testing


def test_config_testing(app):
    assert app.testing


def test_version(app):
    assert conf_mgr.version == karp5.__version__



from karp5.errors import KarpException, KarpAuthenticationError, KarpQueryError, KarpGeneralError

from karp5.server.translator.errors import QueryError, AuthenticationError, BulkifyError


def test_KarpException_default():
    msg = "TEST"
    error = KarpException(msg)

    assert error.message == msg
    assert error.status_code == 400


def test_KarpException_status_code():
    msg = "TEST"
    error = KarpException(msg, status_code=800)

    assert error.message == msg
    assert error.status_code == 800


def test_KarpAuthenticationError_default():
    msg = "TEST"
    error = KarpAuthenticationError(msg)

    expected_message = f"Authentication Exception: {msg}"
    assert error.message == expected_message
    assert error.debug_msg == expected_message
    assert error.status_code == 401


def test_KarpAuthenticationError_status_code():
    msg = "TEST"
    error = KarpAuthenticationError(msg, status_code=800)

    expected_message = f"Authentication Exception: {msg}"
    assert error.message == expected_message
    assert error.debug_msg == expected_message
    assert error.status_code == 800


def test_KarpQueryError_default():
    msg = "TEST"
    error = KarpQueryError(msg)

    assert error.message == f"Query Error: {msg}. No query."


def test_KarpQueryError_query():
    msg = "TEST"
    query = "TEST_QUERY"
    error = KarpQueryError(msg, query=query)

    assert error.message == f"Query Error: {msg}. Query was '{query}'"


def test_KarpGeneralError_no_message():
    msg = None
    error = KarpGeneralError(msg)

    assert error.message == "Error: Unknown error."


def test_KarpGeneralError_message():
    msg = "TEST"
    error = KarpGeneralError(msg)

    assert error.message == f"Error: {msg}"


def test_QueryError():
    msg = "TEST"
    error = QueryError(msg)
    assert error.message == f"malformed query: {msg}"
    assert error.status_code == 400


def test_AuthenticationError():
    msg = "TEST"
    error = AuthenticationError(msg)

    assert error.message == f"Authentication Exception: {msg}"
    assert error.status_code == 401


def test_BulkifyError():
    msg = "TEST"
    bulk_info = "BULK_INFO_TEST"
    error = BulkifyError(msg, bulk_info)

    assert error.message == f"Bulkify error: {msg}\nbulk_info = {bulk_info}"



import datetime
import logging


auto_updates = {}
auto_updates_child = {}

_logger = logging.getLogger("karp5")


def auto_update_child(*modes):
    def wrap(func):
        for mode in modes:
            _logger.debug("add child auto function %s" % func)
            auto_updates_child.setdefault(mode, []).append(func)
        return func

    return wrap


def auto_update(*modes):
    def wrap(func):
        for mode in modes:
            _logger.debug("add auto function %s" % func)
            auto_updates.setdefault(mode, []).append(func)
        return func

    return wrap


def auto_update_document(data_doc, lexiconName, actiontype, user, date):
    _logger.debug("\n\nautoupdate!")
    if not date:
        date = datetime.datetime.now()
    for func in auto_updates.get(lexiconName, []):
        _logger.debug("\n\nautoupdate, apply %s on %s" % (func, data_doc))
        func(data_doc, lexiconName, actiontype, user, date)


def autoupdate_child(child, parent, lexicon, user, date=""):
    _logger.debug("autoupdate child")
    if not date:
        date = datetime.datetime.now()
    for func in auto_updates_child.get(lexicon, []):
        _logger.debug("autoupdate_child, apply %s on %s", func, child)
        func(child, parent, lexicon, user, date)



from unittest import mock

import pytest

from karp5.server.translator.validatejson import validate_json


def test_validate_json_escapes_correct():
    lexicon_id = "Test"
    doc = {"field": "&", "string": "o\u0308"}
    expected = {"field": "&amp;", "string": "ö"}

    validate_json(doc, lexicon_id)

    assert doc == expected


def test_validate_json_respect_no_escape_but_normalizes_string():
    lexicon_id = "test"
    doc = {"field": "&", "string": "o\u0308"}
    expected = {"field": "&", "string": "ö"}

    with mock.patch(
        "karp5.server.translator.validatejson.conf_mgr", lexicons={"test": {"no_escape": True}}
    ):
        validate_json(doc, lexicon_id)

    assert doc == expected


@pytest.mark.parametrize(
    "doc, expected",
    [
        ({"xml": "<tag>"}, {"xml": "<tag>"}),
        ({"xml": "[<tag>]"}, {"xml": "[<tag>]"}),
        ({"xml": {"text": "<tag>"}}, {"xml": {"text": "&lt;tag&gt;"}}),
    ],
)
def test_validate_json_handles_xml_correct_with_escaping(doc, expected):
    lexicon_id = "test"

    with mock.patch(
        "karp5.server.translator.validatejson.conf_mgr", lexicons={"test": {"usedtags": ["tag"]}}
    ):
        validate_json(doc, lexicon_id)

    assert doc == expected


@pytest.mark.parametrize(
    "doc, expected",
    [
        ({"xml": "<tag>"}, {"xml": "<tag>"}),
        ({"xml": "[<tag>]"}, {"xml": "[<tag>]"}),
        ({"xml": {"text": "<tag>"}}, {"xml": {"text": "<tag>"}}),
    ],
)
def test_validate_json_handles_xml_correct_without_escaping(doc, expected):
    lexicon_id = "test"

    with mock.patch(
        "karp5.server.translator.validatejson.conf_mgr",
        lexicons={"test": {"no_escape": True, "usedtags": ["tag"]}},
    ):
        validate_json(doc, lexicon_id)

    assert doc == expected





# startup-scriptet (samma som strix har)

import sys  # noqa: E402

import dotenv  # noqa: E402

dotenv.load_dotenv(dotenv_path=".env", verbose=True)

import karp5  # noqa: E402


if __name__ == "__main__":
    arg = sys.argv[1]
    app = karp5.create_app()
    if arg == "dev":
        app.run(debug=True, port=8081)
    else:
        try:
            port = int(arg)
        except (IndexError, ValueError):
            sys.exit("Usage %s <port>" % sys.argv[0])

        app.run(debug=False, port=port)



import logging
import os
import tempfile
import time
from distutils.util import strtobool

import pytest

from click.testing import CliRunner

import elasticsearch_test

import dotenv

dotenv.load_dotenv(dotenv_path=".env", verbose=True)
os.environ["KARP5_INSTANCE_PATH"] = os.path.join(os.path.dirname(__file__), "data/")
assert os.environ["KARP5_INSTANCE_PATH"] == os.path.join(os.path.dirname(__file__), "data/")
print(os.environ["KARP5_INSTANCE_PATH"])

from karp5 import create_app  # noqa: E402
from karp5.cli import cli as karp5_cli, setup_cli  # noqa: E402
from karp5.config import conf_mgr  # noqa: E402

from karp5.tests._test_config import TestConfig  # pytype: disable=import-error


@pytest.fixture(name="app", scope="session")
def fixture_app():
    app = create_app(TestConfig)

    assert TestConfig.OVERRIDE_ELASTICSEARCH_URL
    return app


# @pytest.fixture(scope="session")
# def real_app():
#     app = create_app()
#
#     return app


from karp5.tests.common_fixtures import (  # pytype: disable=import-error
    fixture_client,
    fixture_cli,
    fixture_cli_w_es,
    fixture_es,
)  # pytype: disable=import-error


@pytest.fixture(name="cli_w_panacea", scope="session")
def fixture_cli_w_panacea(cli_w_es):
    r_create = cli_w_es.create_mode("panacea", "test")
    assert r_create.exit_code == 0

    r_publish = cli_w_es.publish_mode("panacea", "test")
    assert r_publish.exit_code == 0

    time.sleep(1)
    return cli_w_es


@pytest.fixture(name="cli_w_foo", scope="session")
def fixture_cli_w_foo(cli_w_es):
    r_create = cli_w_es.create_mode("foo", "test")
    assert r_create.exit_code == 0

    r_publish = cli_w_es.publish_mode("foo", "test")
    assert r_publish.exit_code == 0

    time.sleep(1)
    return cli_w_es


@pytest.fixture(name="cli_w_large_lex", scope="session")
def fixture_cli_w_large_lex(cli_w_es):
    r_create = cli_w_es.create_mode("large_lex", "test")
    assert r_create.exit_code == 0

    r_publish = cli_w_es.publish_mode("large_lex", "test")
    assert r_publish.exit_code == 0

    time.sleep(1)
    return cli_w_es


@pytest.fixture(name="app_w_panacea", scope="session")
def fixture_app_w_panacea(app, cli_w_panacea):
    if cli_w_panacea is None:
        pytest.skip()
    return app


@pytest.fixture(scope="session")
def client_w_panacea(app_w_panacea):
    return app_w_panacea.test_client()


@pytest.fixture(name="app_w_foo", scope="session")
def fixture_app_w_foo(app, cli_w_foo):
    if cli_w_foo is None:
        pytest.skip()
    return app


@pytest.fixture(scope="session")
def client_w_foo(app_w_foo):
    return app_w_foo.test_client()


@pytest.fixture(name="app_w_large_lex", scope="session")
def fixture_app_w_large_lex(app, cli_w_large_lex):
    if cli_w_large_lex is None:
        pytest.skip()
    return app


@pytest.fixture(scope="session")
def client_w_large_lex(app_w_large_lex):
    return app_w_large_lex.test_client()



"""The command-line interface for Karp."""
import logging
import logging.handlers
import os

import click

import karp5
from karp5.application.actions import mode_actions
from karp5.cli import create_metadata as metadata
from karp5.cli import upload_offline as upload
from karp5.cli import getmapping as gm
from karp5.config import conf_mgr


_logger = logging.getLogger("karp5")


def setup_cli(config=karp5.Config):
    print("Setting up logging")
    karp5.conf_mgr.app_config = config
    logger = logging.getLogger("karp5")
    logger.setLevel(config.LOG_LEVEL)
    formatter = logging.Formatter(fmt=config.LOG_FMT, datefmt=config.LOG_DATEFMT)

    if config.TESTING or config.DEBUG or config.LOG_TO_STDERR:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(config.LOG_LEVEL)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    else:
        log_dir = config.LOG_DIR
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        file_handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(log_dir, "karp5-admin.log"), when="d", interval=1, backupCount=0,
        )
        file_handler.setLevel(config.LOG_LEVEL)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.WARNING)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)


@click.group()
@click.version_option(karp5.get_version(), prog_name="karp5")
def cli():
    if not karp5.conf_mgr.app_config:
        setup_cli()


@cli.command("create_metadata", short_help="Generate fieldmappings.")
def create_metadata():
    """
    Generate 'config/fieldmappings.json'.

    Generate from the 'config/mappings/fieldmappings_*.json' files.
    NOTE: Not needed since version 5.9.0
    """
    outpath = "config/fieldmappings.json"
    metadata.print_all(outpath)


@cli.group(name="mode")
def group_mode():
    """Handle MODES."""


@cli.group(name="lexicon")
def group_lexicon():
    """Handle LEXICONS in MODES."""


@cli.command("create_mode")
@click.argument("mode")
@click.argument("suffix")
def create_mode(mode, suffix):
    upload.create_mode(mode, suffix)
    print("Upload successful")


@group_mode.command("init")
@click.argument("mode", type=str)
@click.argument("suffix", default=None, type=str)
@click.option("--data-dir", "-d", default=None, type=str, help="file to load", metavar="<filename>")
def mode_init(mode, suffix, data_dir):
    """Init MODE.

    Creates the index MODE_SUFFIX.

    If --data is given then the data is loaded from that file,
    otherwise the data is loaded from default location.
    """
    upload.create_mode(mode, suffix, data_dir=data_dir)
    print("Upload successful")


# @group_mode.command("recover")
# @click.argument("mode", type=str)
# @click.argument("suffix", default=None, type=str, required=False)
# @click.option(
#     "--lexicons",
#     default=None,
#     type=str,
#     help="lexicons to recover",
#     metavar="<lexicon1>[,<lexicon2>,...]",
# )
# def mode_recover(mode, suffix, lexicons):
#     """Recover MODE from sql.

#     Copies all entries in sql to the index MODE_SUFFIX.

#     If --lexicons is given, only those lexicons are recovered.
#     """
#     _logger.info("mode_recover called with mode=%s", mode)
#     try:
#         result = mode_actions.recover(
#             mode, suffix=suffix, lexicons=None if not lexicons else lexicons.split(",")
#         )
#         print(result)
#     except Exception as e:
#         _logger.exception(str(e))


@group_lexicon.command("init")
@click.argument("lexicon", type=str)
@click.argument("suffix", default=None, type=str)
@click.option("--data", "-d", default=None, type=str, help="file to load", metavar="<filename>")
def lexicon_init(lexicon, suffix, data):
    """Init the LEXICON.

    Loads the lexicon to the index <mode of LEXICON>_SUFFIX.

    If --data is given then the data is loaded from that file,
    otherwise the data is loaded from default location.
    """
    _logger.debug(
        "lexicon_init(lexicon='%s', suffix='%s', filename='%s')",
        lexicon,
        suffix,
        "None" if data is None else data,
    )
    index = upload.add_lexicon(lexicon, suffix, filename=data)
    print(f"'{lexicon}' was successfully loaded to the index '{index}'")


@group_lexicon.command("recover")
@click.argument("lexicon", type=str)
@click.argument("suffix", default=None, type=str, required=False)
def lexicon_recover(lexicon, suffix):
    """Recovers the LEXICON from sql.

    Loads the lexicon to the index <mode of LEXICON>_SUFFIX.
    """
    _logger.debug(
        "lexicon_recover(lexicon='%s', suffix='%s')", lexicon, suffix,
    )
    mode = conf_mgr.get_lexicon_mode(lexicon)
    result, index = upload.recover(mode, suffix, lexicon)
    if result:
        print(f"'{lexicon}' was successfully loaded to the index '{index}'")
    else:
        print("Something went wrong.")


@cli.command("create_empty_index")
@click.argument("mode")
@click.argument("suffix")
def create_empty_index(mode, suffix):
    upload.create_empty_index(mode, suffix)
    print("Index created successfully")


@cli.command("import_mode")
@click.argument("mode")
@click.argument("suffix")
def import_mode(mode, suffix):
    upload.create_mode(mode, suffix, with_id=True)
    print("Upload successful")


@cli.command("publish_mode")
@click.argument("mode")
@click.argument("suffix")
def publish_mode(mode, suffix):
    """Publish MODE to all modes that contain MODE."""
    upload.publish_mode(mode, suffix)
    print("Published '{mode}_{suffix}' successfully to '{mode}'".format(mode=mode, suffix=suffix))


@cli.command("reindex_alias")
@click.argument("index")
@click.argument("target_suffix")
def reindex_alias(index, target_suffix):
    ret = upload.reindex_alias(index, target_suffix)
    if not ret:
        raise click.ClickException("Something went wrong")


@cli.command("getmapping")
@click.argument("alias")
@click.argument("outfile")
def getmapping(alias, outfile):
    gm.getmapping(alias, outfile)


@cli.command("internalize_lexicon")
@click.argument("mode", nargs=1)
@click.argument("to_upload", nargs=-1)
def internalize_lexicon(mode, to_upload):
    """ Add a lexicon to sql from es.
    Can be done at any time, not noticable to the end user.
    """

    upload.internalize_lexicon(mode, to_upload)
    print("Upload successful")


@cli.command("printlatestversion")
@click.argument("lexicon", metavar="<lexicon>")
@click.option("--output", "-o", default=None, metavar="<path>", help="File to write to.")
def printlatestversion(lexicon, output):
    if output:
        with open(output, "w") as fp:
            upload.printlatestversion(lexicon, fp=fp)
    else:
        upload.printlatestversion(lexicon)


@cli.command("exportlatestversion")
@click.argument("lexicon", metavar="<lexicon>")
@click.option("--output", "-o", default=None, metavar="<path>", help="File to write to.")
def exportlatestversion(lexicon, output):
    if output:
        with open(output, "w") as fp:
            upload.printlatestversion(lexicon, with_id=True, fp=fp)
    else:
        upload.printlatestversion(lexicon, with_id=True)


# Commented since dangerous!
# @cli.command('delete_all')
# def delete_all():
#     """ Deletes all modes.
#     You need to confirm this 5 times before proceding.
#     """
#     print('WARNING: This command will delete all modes.')
#     print('You have to confirm this 5 times.')
#     for i in range(1,6):
#         click.confirm(
#             'Are you sure you want to delete ALL modes ({}/5)?'.format(i),
#             abort=True
#         )
#     upload.delete_all()
#     print('Deletion successful')


# Commented since dangerous!
# @cli.command('delete_mode')
# @click.argument('mode')
# def delete_mode(mode):
#     click.confirm(
#         "Are you sure you want to delete the mode '{}'".format(mode),
#         abort=True
#     )
#     upload.delete_mode(mode)
#     print('Deletion successful')



from karp5.errors import KarpException

""" Class for exceptions generated by the parser modules
"""


class QueryError(KarpException):
    def __init__(self, message, debug_msg=None, status_code=None, payload=None):
        super().__init__(
            "malformed query: " + message,
            status_code=status_code,
            payload=payload,
            debug_msg=debug_msg,
            user_msg=message,
        )

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["message"] = self.message
        return rv


class AuthenticationError(KarpException):
    def __init__(self, message, status_code=None, payload=None):
        super().__init__(
            "Authentication Exception: " + message, status_code=status_code or 401, payload=payload,
        )

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["message"] = self.message
        return rv


class BulkifyError(KarpException):
    def __init__(self, message, bulk_info):
        super().__init__(f"Bulkify error: {message}\nbulk_info = {bulk_info}")



import base64
import hashlib
from unittest.mock import patch, Mock

from karp5.context.auth import std_auth  # pytype: disable=import-error
from karp5.config import conf_mgr


def test_hashlib_md5():
    user = "User"
    pwd = "Pwd"
    secret = "Secret"
    user_pwd_secret = user + pwd + secret

    result = hashlib.md5((user + pwd + secret).encode("utf-8")).hexdigest()
    assert result is not None


def test_check_user_no_user(app):
    with app.test_request_context("/query?q=simple||hej"):
        with patch("karp5.context.auth.std_auth.make_auth_request") as mk_auth_req_mock:
            mk_auth_req_mock.return_value = {
                "authenticated": False,
            }
            result = std_auth.check_user()
            mk_auth_req_mock.assert_called_with(
                conf_mgr.app_config.AUTH_RESOURCES, {"include_open_resources": "true"}
            )
            assert "username" in result
            assert result["username"] == ""


def test_check_user_valid_user(app):
    with patch("karp5.context.auth.std_auth.make_auth_request") as mk_auth_req_mock:
        username = "valid_user"
        password = "pwd"
        creds = base64.b64encode(b"valid_user:pwd").decode("utf-8")
        with app.test_request_context(
            "/query?q=simple||hej", headers={"Authorization": "Basic " + creds}
        ):
            mk_auth_req_mock.return_value = {
                "authenticated": True,
                "permitted_resources": {
                    "lexica": {"a": {"read": True, "write": True, "admin": True}}
                },
            }
            result = std_auth.check_user()
            mk_auth_req_mock.assert_called_with(
                conf_mgr.app_config.AUTH_SERVER,
                {
                    "include_open_resources": "true",
                    "username": username,
                    "password": password,
                    "checksum": "84f0163c1b904d544dd0e1d9168bd58b",
                },
            )
            assert "username" in result
            assert result["username"] == username
            assert result["authenticated"]
            assert "lexicon_list" in result
            assert "a" in result["lexicon_list"]
            assert isinstance(result["lexicon_list"]["a"], dict)
            assert "auth_response" in result


def test_check_user_invalid_user(app):
    with patch("karp5.context.auth.std_auth.make_auth_request") as mk_auth_req_mock:
        username = "invalid_user"
        password = "pwd"
        creds = base64.b64encode(b"invalid_user:pwd").decode("utf-8")
        with app.test_request_context(
            "/query?q=simple||hej", headers={"Authorization": "Basic " + creds}
        ):
            mk_auth_req_mock.return_value = {
                "authenticated": False,
                "permitted_resources": {
                    "lexica": {"open": {"read": True, "write": False, "admin": False}}
                },
            }
            result = std_auth.check_user()
            mk_auth_req_mock.assert_called_with(
                conf_mgr.app_config.AUTH_SERVER,
                {
                    "include_open_resources": "true",
                    "username": username,
                    "password": password,
                    "checksum": "8832a3ca67946278bcc74e138cd0718d",
                },
            )
            assert "username" in result
            assert result["username"] == username
            assert not result["authenticated"]
            assert "lexicon_list" in result
            assert "open" in result["lexicon_list"]
            assert isinstance(result["lexicon_list"]["open"], dict)
            assert result["lexicon_list"]["open"]["read"]
            assert not result["lexicon_list"]["open"]["write"]
            assert not result["lexicon_list"]["open"]["admin"]
            assert "auth_response" in result



""" Methods for updating the database,
    including deletion and creation of indices
"""


import datetime
import logging
from elasticsearch import helpers as eshelpers
from elasticsearch import exceptions as esExceptions
from flask import request, jsonify
from json import dumps
from karp5 import errors
from karp5.config import mgr as conf_mgr
from karp5.context import auth
import karp5.server.helper.helpers as helpers
import karp5.server.translator.validatejson as validate
from karp5.document import auto_update_document
from karp5.document import autoupdate_child
from karp5.document import doc_to_es


_logger = logging.getLogger("karp5")


def checkuser():
    """ Shows which lexica the user may edit """
    authdict, permitted = auth.validate_user(mode="verbose")
    return jsonify(permitted)


def delete_entry(lexicon, _id, sql=False, live=True, suggestion=False):
    # delete by id
    try:
        msg = request.args.get("message", "removed")
        es, index, typ = helpers.get_update_index(lexicon, suggestion=suggestion)
        ans_entry = es.get(index=index, doc_type=typ, id=_id)
        lexiconName = ans_entry["_source"]["lexiconName"]
        helpers.check_lexiconName(lexicon, lexiconName, _id, "delete")

        authdict, permitted = auth.validate_user()
        if lexiconName not in permitted:
            raise errors.KarpAuthenticationError(
                "You are not allowed to modify "
                "the lexicon %s, only %s" % (lexiconName, permitted)
            )

        # doc_type must be set
        ans = es.delete(index=index, doc_type=typ, id=_id)
        db_loaded, db_error = 0, ""
        if sql:
            # _logger.debug("Delete " + msg)
            _logger.debug("delete from sql.\nmsg %s\nans_entry %s", msg, ans_entry)
            db_loaded, db_error = update_db(
                _id, ans_entry["_source"], helpers.get_user(), msg, lexiconName, status="removed",
            )
            _logger.debug("updated db %s %s", db_loaded, db_error)

        if db_error:
            raise errors.KarpDbError(db_error)

    except eshelpers.BulkIndexError as e:
        # BulkIndexException is thrown for other parse errors
        # This exception has errors instead of error
        handle_update_error(e, {"id": _id}, helpers.get_user(), "delete")
        err = [er["create"]["error"] for er in e.errors]
        raise errors.KarpElasticSearchError("Error during deletion %s.\n" % "\n".join(err))

    except (esExceptions.TransportError, esExceptions.RequestError) as e:
        # elasticsearch-py throws errors (TransportError)
        # for invalid (empty) objects
        handle_update_error(e, {"id": _id}, helpers.get_user(), "delete")
        err = [e.error]
        raise errors.KarpElasticSearchError(
            "Error during deletion. " "Message: %s.\n" % "\n".join(err)
        )

    except Exception as e:
        handle_update_error(e, {"id": _id}, helpers.get_user(), "delete")
        err = ["Oops, an unpredicted error", str(e), "Document not deleted"]
        raise errors.KarpGeneralError("Document not deleted", debug_msg=" ".join(err))

    jsonans = {"es_loaded": 1, "sql_loaded": db_loaded, "es_ans": ans}
    if db_error:
        _logger.debug(db_error)
    if live:
        return jsonify(jsonans)
    else:
        return jsonans


def get_lexname(data):
    names = set()
    for item in data:
        names.add(item["_source"]["lexiconName"])
    return names


def get_sql_and_es_doc(doc):
    if "sql" in doc and "es" in doc:
        return doc["sql"], doc["es"]
    else:
        return doc, doc


def update_doc(lexicon, _id, data=None, live=True):
    """ Updates a posted document in the index 'index' with type 'typ'.
        The document must contain a field called 'doc' with
        the information to be sent.
        The fields 'version' and 'message' are optional.
    """
    # send user name and password,
    # {'doc' : es_doc, 'version' : last version, 'message' : update message}
    authdict, permitted = auth.validate_user()

    if data is None:
        data = helpers.read_data()

    try:
        index, typ = conf_mgr.get_lexicon_index(lexicon)
        es = conf_mgr.elastic(lexicon=lexicon)
        origin = es.get(index=index, doc_type=typ, id=_id)
    except Exception as e:
        _logger.warning("Looking for entry at the wrong place:")
        _logger.exception(e)
        msg = "The entry %s in lexicon %s was not found" % (_id, lexicon)
        raise errors.KarpElasticSearchError(msg, debug_msg=msg + " in lexicon " + lexicon)

    if "lexiconName" in origin["_source"]:
        lexiconName = origin["_source"]["lexiconName"]
    else:
        lexiconName = origin["_source"]["_lexiconName"]
    # aborts if mismatch
    helpers.check_lexiconName(lexicon, lexiconName, _id, "update")
    data_doc = data.get("doc") or data.get("_source")
    version = data.get("version")
    msg = data["message"]

    if lexicon not in permitted:
        raise errors.KarpAuthenticationError(
            "You are not allowed to modify the " "lexicon %s, only %s" % (lexicon, permitted),
            status_code=403,
        )

    validate.validate_json(data_doc, lexicon)
    date = datetime.datetime.now()
    user = helpers.get_user()
    auto_update_document(data_doc, lexicon, "update", user, date)
    es_doc = doc_to_es(data_doc, lexiconName, "update", user, date)

    try:
        if version is not None and version != -1:
            ans = es.index(
                index=index, doc_type=typ, id=_id, version=version, body=es_doc, op_type="index",
            )
        else:
            ans = es.index(index=index, doc_type=typ, id=_id, body=es_doc, op_type="index")

    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        # Transport error might be version conflict
        _logger.exception(e)
        _logger.debug("index: %s, type: %s, id: %s", index, typ, _id)
        handle_update_error(e, {"id": _id, "data": data}, user, "update")
        raise errors.KarpElasticSearchError("Error during update. Message: %s.\n" % str(e))
    except Exception as e:
        handle_update_error(e, {"id": _id, "data": data}, user, "update")
        raise errors.KarpElasticSearchError("Unexpected error during update.")

    db_loaded, db_error = update_db(
        _id,
        data_doc,
        user,
        msg,
        lexiconName,
        status="changed",
        date=date,
        version=ans.get("_version", version),
    )

    jsonans = {"es_loaded": 1, "sql_loaded": db_loaded, "es_ans": ans}
    if db_error:
        # TODO: handle partial success.
        # ans = es.delete(index=index, doc_type=typ, id=_id)
        _logger.debug(db_error)
    if live:
        return jsonify(jsonans)
    else:
        return jsonans


def update_db(_id, doc, user, msg, lex, status="", version="", suggestion="", date=""):
    """ Inserts the document doc into the sql data base, using its id and
        the current date and time as a key. The user name and a message is
        also saved.
        Returns a tuple (loaded, error), where 'loaded' is 1 if the operation
        succeeded and 0 otherwise and 'error' is the error message given, if
        any.
    """
    from karp5.dbhandler.dbhandler import update

    return update(
        _id,
        dumps(doc),
        user,
        msg,
        lex,
        status=status,
        version=version,
        suggestion_id=suggestion,
        date=date,
    )


def modify_db(_id, lexicon, msg, status, origid=""):
    """ Updates a document in the suggestion sql data base, using its id.
        Returns a tuple (loaded, error), where 'loaded' is 1 if the operation
        succeeded and 0 otherwise and 'error' is the error message given, if
        any.
    """
    from karp5.dbhandler.dbhandler import modifysuggestion

    return modifysuggestion(_id, lexicon, msg=msg, status=status, origid=origid)


# TODO Is this used?
def add_multi_doc(lexicon, index=""):
    import karp5.dbhandler.dbhandler as db

    data = helpers.read_data()
    documents = data.get("doc", "") or data.get("_source")
    message = data["message"]
    es, index, typ = helpers.get_update_index(lexicon)
    # validate that the user may update the lexica
    authdict, permitted = auth.validate_user()
    if lexicon not in permitted:
        errstr = "You are not allowed to modify the lexicon %s"
        raise errors.KarpAuthenticationError(errstr % lexicon, status_code=403)
    user = helpers.get_user()
    try:
        bulk, sql_bulk, ids = [], [], []
        ok = 0
        for doc in documents:
            doc["lexiconName"] = lexicon
            validate.validate_json(doc, lexicon)
            date = datetime.datetime.now()
            _logger.debug("\n\nWill go to autoupdate\n\n")
            auto_update_document(doc, lexicon, "add", user, date)
            bulk.append({"_index": index, "_type": typ, "_source": doc})

        for ix, res in enumerate(eshelpers.streaming_bulk(es, bulk)):
            _id = res[1].get("index").get("_id")
            source = bulk[ix].get("_source")
            if isinstance(source, dict):
                source = dumps(source)
            sql_bulk.append((_id, source, user, "multi add - %s" % message, lexicon, "imported"))
            ids.append(_id)
            ok += 1

    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        handle_update_error(e, data, user, "add")
        raise errors.KarpElasticSearchError("Error during update. Message: %s.\n" % str(e))

    db_loaded, db_error = db.update_bulk(lexicon, sql_bulk)
    if db_error:
        _logger.debug(db_error)

    jsonans = {
        "es_loaded": ok,
        "sql_loaded": db_loaded,
        "suggestion": False,
        "ids": ids,
    }
    return jsonify(jsonans)


def add_doc(lexicon, index="", _id=None, suggestion=False, data=None, live=True):
    """ Adds an entry to the index 'index' with type 'typ' in ES and sql.
        The post data must contain a field called 'doc' with the information to
        be sent.
        The fields 'version' and 'message' are optional.
    """
    if data is None or not data:
        data = helpers.read_data()

    data_doc = data.get("doc", "") or data.get("_source")
    message = data["message"]
    version = data.get("version", -1)
    es, index, typ = helpers.get_update_index(lexicon, suggestion=suggestion)
    lexiconName = lexicon or data_doc.get("lexiconName", "")
    helpers.check_lexiconName(lexicon, lexiconName, "add", _id)

    # lexiconOrder = data_doc.get("lexiconOrder", None)
    if not lexiconName:
        raise errors.KarpParsingError("The field lexiconName is empty, " "although it is required.")

    if suggestion:
        orgin_id = _id or True  # save as reference in db
        _id = ""  # add as new to the suggestion index
        status = "waiting"
        user = data["user"]
    else:
        # validate that the user may update the lexica
        authdict, permitted = auth.validate_user()
        if lexiconName not in permitted:
            errstr = "You are not allowed to modify the lexicon %s"
            raise errors.KarpAuthenticationError(errstr % lexiconName, status_code=403)

        orgin_id = ""  # not a suggestion
        user = helpers.get_user()
        status = "added"
    try:
        validate.validate_json(data_doc, lexicon)

        date = datetime.datetime.now()
        # Make general updates
        auto_update_document(data_doc, lexiconName, "add", user, date)
        # Possibly make index specific updates
        es_doc = doc_to_es(data_doc, lexiconName, "add", user, date)
        ans = es.index(index=index, doc_type=typ, body=es_doc, id=_id)
        _id = ans.get("_id")
        version = ans.get("_version", -1)
        db_loaded, db_error = update_db(
            _id,
            data_doc,
            user,
            message,
            lexiconName,
            version=version,
            suggestion=orgin_id,
            status=status,
            date=date,
        )

    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        handle_update_error(e, data, user, "add")
        raise errors.KarpElasticSearchError("Error during update. Message: %s.\n" % str(e))
    except Exception as e:
        raise errors.KarpGeneralError(str(e))

    jsonans = {
        "es_loaded": 1,
        "sql_loaded": db_loaded,
        "es_ans": ans,
        "suggestion": suggestion,
        "id": _id,
    }
    if db_error:
        _logger.debug(db_error)
    if live:
        return jsonify(jsonans)
    else:
        return jsonans


def add_child(lexicon, parentid, suggestion=False):

    es, index, typ = helpers.get_update_index(lexicon, suggestion=suggestion)
    parent = es.get(index=index, doc_type=typ, id=parentid).get("_source")

    data = helpers.read_data()
    data_doc = data.get("doc", "") or data.get("_source")
    msg = data["message"]
    user = helpers.get_user()

    # add child to parent
    autoupdate_child(data_doc, parent, lexicon, user, "")
    parent_doc = {"doc": parent, "user": user, "message": msg}
    _logger.debug("save parent %s", parent_doc)
    _logger.debug("save child %s", parent_doc)
    parres = update_doc(lexicon, parentid, data=parent_doc, live=False)
    # add child to lexicon
    childres = add_doc(lexicon, index=index, data=data, live=False)
    return jsonify({"parent": parres, "child": childres})


def handle_update_error(error, data, user, action):
    """ Sends emails to admins about updates that failed
    """
    from karp5.dbhandler.dbhandler import handle_error

    return handle_error(error, user, "Action: " + action, data)


def send_notification(user, message, _id, status):
    import re

    if re.match(r"^[^@]+@[^@]+\.[^@]+$", user):
        import karp5.dbhandler.emailsender as sender

        msg = 'Your suggestion with id %s has been %s with the message:\n"%s"' % (
            _id,
            status,
            message,
        )
        subject = "Karp: The status of your suggestion has been changed"
        sender.send_notification(user, subject, msg)



import logging
from typing import List, Tuple

from karp5.context.auth import std_auth

_logger = logging.getLogger("karp5")

check_user = std_auth.check_user


def validate_user(mode: str = "write") -> Tuple[bool, List]:
    """Authenticates a user against an authentication server.
       Returns a dictionary with permitted resources for the user
    """
    # If mode is read and no user info is provided, just get all open lexicons
    # if mode is read and user info is provided, return all lexicons that may
    # be seen by this user
    # if mode is not read, do as before
    # new auth has Lexicon in uppercase, singular

    if mode == "verbose":
        auth = check_user(force_lookup=True)
        return auth.get("authenticated", False), auth.get("auth_response", [])

    user_auth = check_user()
    auth_response = user_auth["authenticated"]

    allowed = []
    for lex, val in user_auth["lexicon_list"].items():
        if val[mode]:
            allowed.append(lex)

    return auth_response, allowed


def init(auth_name: str):
    global check_user
    if auth_name == "dummy":
        _logger.warning("Using dummy auth module.")
        from . import dummy_auth  # pytype: disable=import-error

        check_user = dummy_auth.check_user
    else:

        check_user = std_auth.check_user



from unittest import mock

import pytest

import elasticsearch_dsl as es_dsl

from karp5.domain.services import search_service


@pytest.mark.parametrize(
    "from_,size,expected_calls",
    [
        # (None, None, [mock.call.extra(from_=0, size=0), mock.call.execute().to_dict()]),
        (
            0,
            25,
            [
                mock.call.extra(size=25),
                mock.call.extra().execute(),
                mock.call.extra().execute().to_dict(),
            ],
        ),
        (
            0,
            25,
            [
                mock.call.extra(from_=0, size=25),
                mock.call.extra().execute(),
                mock.call.extra().execute().to_dict(),
            ],
        ),
    ],
)
def test_execute_query_execute(from_, size, expected_calls):
    scan_limit = 1000
    es_search_mock = mock.Mock(name="es_search", spec=es_dsl.Search)
    # es_search_params_mock = mock.Mock(name="es_search_params", spec=es_dsl.Search)
    # es_search_mock.params.return_value = es_search_params_mock
    # es_search_params_mock.scan.return_value = []
    with mock.patch("karp5.config.conf_mgr") as conf_mgr_mock:
        conf_mgr_mock.app_config.return_value.SCAN_LIMIT = scan_limit
        # conf_mgr_mock.__getitem__.return_value = conf_mgr_mock
        search_service.execute_query(es_search_mock, from_=from_, size=size)

    # assert es_search_mock.mock_calls == expected_calls


@pytest.mark.parametrize(
    "from_,size,expected_calls",
    [
        (
            5,
            1200,
            [
                mock.call.extra(from_=0, size=0),
                mock.call.extra().execute(),
                mock.call.params(preserve_order=True, scroll="5m"),
            ],
        ),
    ],
)
def test_execute_query_scan(from_, size, expected_calls):
    scan_limit = 1000
    es_search_mock = mock.Mock(name="es_search", spec=es_dsl.Search)
    es_search_params_mock = mock.Mock(name="es_search_params", spec=es_dsl.Search)
    es_search_mock.params.return_value = es_search_params_mock
    es_search_params_mock.scan.return_value = []
    with mock.patch("karp5.config.conf_mgr") as conf_mgr_mock:
        conf_mgr_mock.app_config.return_value.SCAN_LIMIT = scan_limit
        # conf_mgr_mock.__getitem__.return_value = conf_mgr_mock
        search_service.execute_query(es_search_mock, from_=from_, size=size)

    # assert es_search_mock.mock_calls == expected_calls
    # assert es_search_params_mock.mock_calls == [mock.call.scan()]



from karp5.server import searching

from karp5.tests.util import get_json


def test_autocomplete_with_q(client_w_panacea):
    result = get_json(client_w_panacea, "autocomplete?q=sig")
    assert result["hits"]["total"] == 0
    # https://ws.spraakbanken.gu.se/ws/karp/v5/autocomplete?multi=kasta,docka&resource=saldom&mode=external
    # https://ws.spraakbanken.gu.se/ws/karp/v5/autocomplete?q=kasus&resource=saldom,dalin,hellqvist
    # https://ws.spraakbanken.gu.se/ws/karp/v5/autocomplete?q=kasta&resource=saldom


def test_getcontext_panacea(client_w_panacea):
    result = get_json(client_w_panacea, "getcontext/panacea")
    center_source = {
        "english": [
            {
                "corpus_prob": 4.89859255463686e-06,
                "lemma_english": "suppression",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 3.8369042032346446e-05,
            },
            {
                "corpus_prob": 0.0,
                "lemma_english": "mining",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 0.0,
            },
            {
                "corpus_prob": 3.5721305315087224e-05,
                "lemma_english": "removal",
                "package_prob": 0.1393939393939394,
                "pos_english": "No",
                "target_prob": 0.0002100906287505403,
            },
            {
                "corpus_prob": 7.780117586776189e-06,
                "lemma_english": "degradation",
                "package_prob": 0.18484848484848485,
                "pos_english": "No",
                "target_prob": 3.622252219836902e-05,
            },
            {
                "corpus_prob": 0.0,
                "lemma_english": "quarrying",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 0.0,
            },
            {
                "corpus_prob": 1.6607189268563003e-05,
                "lemma_english": "breakdown",
                "package_prob": 0.015151515151515152,
                "pos_english": "No",
                "target_prob": 2.4148348132246012e-05,
            },
            {
                "corpus_prob": 0.0001688573668833647,
                "lemma_english": "reduction",
                "package_prob": 0.6545454545454545,
                "pos_english": "No",
                "target_prob": 0.0007480621621411321,
            },
            {
                "corpus_prob": 0.0,
                "lemma_english": "dismantling",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 0.0,
            },
            {
                "corpus_prob": 2.4973216945207523e-07,
                "lemma_english": "run-down",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 0.0,
            },
            {
                "corpus_prob": 2.3052200257114636e-07,
                "lemma_english": "dismantlement",
                "package_prob": 0.006060606060606061,
                "pos_english": "No",
                "target_prob": 5.366299584943559e-07,
            },
            {
                "corpus_prob": 4.89859255463686e-06,
                "lemma_english": "collapse",
                "package_prob": 0.0,
                "pos_english": "No",
                "target_prob": 0.00017198990169744105,
            },
        ],
        "lemma_german": "Abbau",
        "lexiconName": "panacea",
        "lexiconOrder": 0,
        "pos_german": "No",
    }
    post = [
        {
            "_id": "qLsaOXABkxee4U0WZNC1",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [{"lemma_english": "image"}, {"lemma_english": "copy"}],
                "lemma_german": "Abbild",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abbild"],
        },
        {
            "_id": "yrsaOXABkxee4U0WZdFo",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "figure"},
                    {"lemma_english": "projection"},
                    {"lemma_english": "illustration"},
                    {"lemma_english": "image"},
                    {"lemma_english": "mapping"},
                ],
                "lemma_german": "Abbildung",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abbildung"],
        },
        {
            "_id": "-bsaOXABkxee4U0WYMj7",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "termination"},
                    {"lemma_english": "disconnection"},
                    {"lemma_english": "demolition"},
                    {"lemma_english": "abortion"},
                    {"lemma_english": "breaking-off"},
                    {"lemma_english": "abort"},
                ],
                "lemma_german": "Abbruch",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abbruch"],
        },
        {
            "_id": "V7saOXABkxee4U0WXsJb",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "cover"},
                    {"lemma_english": "coverage"},
                    {"lemma_english": "covering"},
                ],
                "lemma_german": "Abdeckung",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abdeckung"],
        },
        {
            "_id": "lbsaOXABkxee4U0WZM4c",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "imprint"},
                    {"lemma_english": "impression"},
                    {"lemma_english": "printing"},
                    {"lemma_english": "print"},
                    {"lemma_english": "copy"},
                ],
                "lemma_german": "Abdruck",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abdruck"],
        },
        {
            "_id": "LbsaOXABkxee4U0WYMZl",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [{"lemma_english": "dinner"}, {"lemma_english": "supper"}],
                "lemma_german": "Abendessen",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abendessen"],
        },
        {
            "_id": "ebsaOXABkxee4U0WX8SF",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "adventure"},
                    {"lemma_english": "affair"},
                    {"lemma_english": "venture"},
                ],
                "lemma_german": "Abenteuer",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abenteuer"],
        },
        {
            "_id": "AbsaOXABkxee4U0WXMHw",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [{"lemma_english": "deprivation"}, {"lemma_english": "denial"}],
                "lemma_german": "Aberkennung",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Aberkennung"],
        },
        {
            "_id": "-7saOXABkxee4U0WZdJp",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [
                    {"lemma_english": "refuse"},
                    {"lemma_english": "waste"},
                    {"lemma_english": "offal"},
                    {"lemma_english": "rubbish"},
                    {"lemma_english": "drop"},
                    {"lemma_english": "trash"},
                    {"lemma_english": "litter"},
                    {"lemma_english": "scrap"},
                    {"lemma_english": "discard"},
                    {"lemma_english": "garbage"},
                    {"lemma_english": "release"},
                ],
                "lemma_german": "Abfall",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abfall"],
        },
        {
            "_id": "KLsaOXABkxee4U0WWb_M",
            "_index": "panacea_test_upload_02",
            "_score": None,
            "_source": {
                "english": [{"lemma_english": "intercept"}, {"lemma_english": "interception"}],
                "lemma_german": "Abfangen",
                "lexiconName": "panacea",
            },
            "_type": "lexicalentry",
            "sort": ["Abfangen"],
        },
    ]
    pre = []
    assert result["pre"] == pre
    assert result["center"]["_source"] == center_source
    for post_entry, expected in zip(result["post"], post):
        assert post_entry["_source"] == expected["_source"]



from karp5.config.configmanager import ConfigManager
import os

import pytest

from karp5 import conf_mgr, config
from karp5.instance_info import get_instance_path
from karp5.config import errors


def test_override_elastic_url():
    mgr = config.ConfigManager("path")
    elastic_url = "test.elastic.url"

    for _, mode_settings in mgr.modes.items():
        assert "elastic_url" in mode_settings
        assert mode_settings["elastic_url"] != elastic_url

    mgr.override_elastic_url(elastic_url)

    for _, mode_settings in mgr.modes.items():
        assert "elastic_url" in mode_settings
        assert mode_settings["elastic_url"] == elastic_url


def test_config_for_test(app):
    assert conf_mgr.app_config.OVERRIDE_ELASTICSEARCH_URL
    if os.environ.get("KARP5_DBPASS"):
        assert conf_mgr.app_config.DATABASE_BASEURL.startswith("mysql")
    else:
        assert conf_mgr.app_config.DATABASE_BASEURL.startswith("sqlite")
    assert conf_mgr.get_mode_sql("panacea") is not None
    assert conf_mgr.get_mode_sql("default") is not None
    assert conf_mgr.get_mode_sql("karp") is not None


# def test_real_config(real_app):
#     assert config.mgr.app_config.DATABASE_BASEURL.startswith('mysql')


def test_get_mapping(app):
    mapping = config.mgr.get_mapping("panacea")
    assert isinstance(mapping, str)

    with pytest.raises(errors.KarpConfigException) as e:
        mapping = config.mgr.get_mapping("not-existing")
    assert "Can't find mappingconf for index 'not-existing'" in str(e.value)


def test_default_filename(app):
    lexicon = "foo"
    filename = conf_mgr.default_filename(lexicon)

    assert filename == os.path.join(os.path.abspath(get_instance_path()), "data", "foo", "foo.json")


@pytest.mark.parametrize("mode,facit", [("panacea", ["panacea", "karp", "panacea_links"])])
def test_get_modes_that_include_mode(app, mode, facit):
    modes = config.mgr.get_modes_that_include_mode(mode)
    assert len(modes) == len(facit)
    for m in modes:
        assert m in facit
    for f in facit:
        assert f in modes


@pytest.mark.parametrize("mode,expected", [("karp", None), ("foo", {"status": "ok"}),])
def test_filter_for_unauth_user(app, mode, expected):
    assert conf_mgr.filter_for_unauth_user(mode) == expected


def test_lookup_multiple_score(app):
    mode = "panacea"
    result = conf_mgr.lookup_multiple("_score", mode)
    expected = ["_score"]

    assert result == expected


def test_version_default():
    mgr = config.ConfigManager("path")

    assert mgr.version == ""


def test_update_version():
    mgr = config.ConfigManager("path")

    version = "x.y.z"
    mgr.update_version(version)

    assert mgr.version == version




from karp5.domain.model.mode import Mode

MODE_DEFAULTS = {
    "elastic_url": ["http://elasticsearch:9200"],
    "sql": "karp",
    "type": "lexicalentry",
    "secret_fields": ["lastmodifiedBy"],
    "src": "",
    "suggestionalias": "karpsuggestion",
    "sort_by": ["lexiconOrder", "_score", "baseform.sort"],
    "head_sort_field": ["lexiconOrder"],
    "autocomplete_field": ["baseform"],
    "minientry_fields": ["lexiconName", "baseform", "baseform_en"],
    "statistics_buckets": ["lexiconName", "pos"],
    "boosts": ["baseform", "baseform_en"],
}


def test_mode_create_empty():
    mode = Mode(id="id", type="type", indexalias="indexalias", suggestionalias="suggestionalias")

    assert mode.id == "id"
    assert mode.elastic_url == []
    assert mode.sql is None
    assert mode.type == "type"
    assert mode.secret_fields == []
    assert mode.src is None
    assert mode.suggestionalias == "suggestionalias"
    assert mode.sort_by == []
    assert mode.head_sort_field == []
    assert mode.autocomplete_field == []
    assert mode.minientry_fields == []
    assert mode.statistics_buckets == []
    assert mode.boosts == []
    assert mode.is_index is False
    assert mode.indexalias == "indexalias"
    assert mode.groups == []
    assert mode.filter_for_unauth_user is None


def test_mode_create_from_dict_with_defaults():
    config = {
        "is_index": True,
        "indexalias": "large_lex",
        "secret_fields": [],
        "src": "",
        "suggestionalias": "barsuggestion",
        "sort_by": ["lexiconOrder", "_score", "foo.sort"],
        "head_sort_field": ["lexiconOrder"],
        "autocomplete_field": ["foo"],
        "minientry_fields": ["lexiconName", "foo"],
        "statistics_buckets": ["lexiconName", "foo"],
        "boosts": ["foo"],
    }
    mode = Mode.from_mapping(id="large_lex", mapping=config, defaults=MODE_DEFAULTS)

    assert mode.id == "large_lex"
    assert mode.elastic_url == MODE_DEFAULTS["elastic_url"]
    assert mode.sql == MODE_DEFAULTS["sql"]
    assert mode.type == MODE_DEFAULTS["type"]
    assert mode.secret_fields == config["secret_fields"]
    assert mode.src is None
    assert mode.suggestionalias == config["suggestionalias"]
    assert mode.sort_by == config["sort_by"]
    assert mode.head_sort_field == config["head_sort_field"]
    assert mode.autocomplete_field == config["autocomplete_field"]
    assert mode.minientry_fields == config["minientry_fields"]
    assert mode.statistics_buckets == config["statistics_buckets"]
    assert mode.boosts == config["boosts"]
    assert mode.is_index == config["is_index"]
    assert mode.indexalias == config["indexalias"]
    assert mode.groups == []
    assert mode.filter_for_unauth_user is None


def test_mode_create_from_dict_without_defaults():
    config = {
        "is_index": True,
        "indexalias": "large_lex",
        "secret_fields": [],
        "src": "",
        "suggestionalias": "barsuggestion",
        "sort_by": ["lexiconOrder", "_score", "foo.sort"],
        "head_sort_field": ["lexiconOrder"],
        "autocomplete_field": ["foo"],
        "minientry_fields": ["lexiconName", "foo"],
        "statistics_buckets": ["lexiconName", "foo"],
        "boosts": ["foo"],
        "type": "entry",
    }
    mode = Mode.from_mapping(id="large_lex", mapping=config)

    assert mode.id == "large_lex"
    assert mode.elastic_url == []
    assert mode.sql is None
    assert mode.type == config["type"]
    assert mode.secret_fields == config["secret_fields"]
    assert mode.src is None
    assert mode.suggestionalias == config["suggestionalias"]
    assert mode.sort_by == config["sort_by"]
    assert mode.head_sort_field == config["head_sort_field"]
    assert mode.autocomplete_field == config["autocomplete_field"]
    assert mode.minientry_fields == config["minientry_fields"]
    assert mode.statistics_buckets == config["statistics_buckets"]
    assert mode.boosts == config["boosts"]
    assert mode.is_index == config["is_index"]
    assert mode.indexalias == config["indexalias"]
    assert mode.groups == []
    assert mode.filter_for_unauth_user is None


def test_mode_create_changing_default_value_dont_changes_mode_defaults():
    config = {
        "is_index": True,
        "indexalias": "large_lex",
        "secret_fields": [],
        "src": "",
        "suggestionalias": "barsuggestion",
        "sort_by": ["lexiconOrder", "_score", "foo.sort"],
        "head_sort_field": ["lexiconOrder"],
        "autocomplete_field": ["foo"],
        "minientry_fields": ["lexiconName", "foo"],
        "statistics_buckets": ["lexiconName", "foo"],
        "boosts": ["foo"],
    }
    mode = Mode.from_mapping(id="large_lex", mapping=config, defaults=MODE_DEFAULTS)

    assert mode.elastic_url == MODE_DEFAULTS["elastic_url"]

    mode.elastic_url.append("added")

    assert mode.elastic_url == ["http://elasticsearch:9200", "added"]
    assert MODE_DEFAULTS["elastic_url"] == ["http://elasticsearch:9200"]


def test_mode_create_changing_mode_defaults_changes_values():
    config = {
        "is_index": True,
        "indexalias": "large_lex",
        "secret_fields": [],
        "src": "",
        "suggestionalias": "barsuggestion",
        "sort_by": ["lexiconOrder", "_score", "foo.sort"],
        "head_sort_field": ["lexiconOrder"],
        "autocomplete_field": ["foo"],
        "minientry_fields": ["lexiconName", "foo"],
        "statistics_buckets": ["lexiconName", "foo"],
        "boosts": ["foo"],
    }
    mode = Mode.from_mapping(id="large_lex", mapping=config, defaults=MODE_DEFAULTS)

    assert mode.elastic_url == MODE_DEFAULTS["elastic_url"]

    mode.secret_fields.append("added")

    assert mode.secret_fields == ["added"]
    assert config["secret_fields"] == ["added"]



import pytest


@pytest.mark.xfail(reason="fails though the command works live.")
def test_cli_recover(cli_w_foo):
    mode = "foo"
    lexicon = "foo"
    suffix = "test_cli_recover"

    r = cli_w_foo.recover_mode(mode, lexicons=[lexicon], suffix=suffix)

    assert r.exit_code == 0



"""[summary]
"""


from distutils.util import strtobool

import json
import logging
import os

from karp5 import instance_info


with open(os.path.join(instance_info.get_instance_path(), "config/config.json")) as fp:
    _CONFIG = json.load(fp)

ENV_DBPASS = os.environ.get("KARP5_DBPASS")
ENV_AUTH_SECRET = os.environ.get("KARP5_AUTH_SECRET")
ENV_SETUP_SECRET_KEY = os.environ.get("KARP5_SETUP_SECRET_KEY")


class Config:
    """[summary]"""

    DEBUG = False
    TESTING = False
    LOG_LEVEL = getattr(logging, _CONFIG["DEBUG"]["DEBUGLEVEL"].upper(), logging.WARNING)
    # LOG_LEVEL = debug_str_to_int(_CONFIG['DEBUG']['DEBUGLEVEL'])
    LOG_FMT = _CONFIG["DEBUG"].get("LOGFMT")
    LOG_DIR = _CONFIG["DEBUG"].get("LOGDIR")
    LOG_DATEFMT = _CONFIG["DEBUG"].get("DATEFMT")
    LOG_TO_STDERR = _CONFIG["DEBUG"].get("DEBUG_TO_STDERR")
    SECRET_KEY = ENV_SETUP_SECRET_KEY or _CONFIG["SETUP"]["SECRET_KEY"]
    ABSOLUTE_PATH = _CONFIG["SETUP"]["ABSOLUTE_PATH"]
    BACKEND_URL = _CONFIG["SETUP"].get("BACKEND_URL") or "https://ws.spraakbanken.gu.se/ws/karp/v5"
    SCRIPT_PATH = _CONFIG["SETUP"]["SCRIPT_PATH"]
    SENDER_EMAIL = _CONFIG["DB"]["SENDER_EMAIL"]
    SMTP_SERVER = _CONFIG["DB"].get("SMTP_SERVER")
    AUTH_RESOURCES = _CONFIG["AUTH"]["AUTH_RESOURCES"]
    AUTH_SECRET = ENV_AUTH_SECRET or _CONFIG["AUTH"]["AUTH_SECRET"]
    AUTH_SERVER = _CONFIG["AUTH"]["AUTH_SERVER"]
    ADMIN_EMAILS = _CONFIG["DB"].get("ADMIN_EMAILS")
    MAX_PAGE = _CONFIG["SETUP"]["MAX_PAGE"]
    MINIENTRY_PAGE = _CONFIG["SETUP"]["MINIENTRY_PAGE"]
    SCAN_LIMIT = _CONFIG["SETUP"]["SCAN_LIMIT"]
    ELASTICSEARCH_URL = os.environ.get("KARP5_ELASTICSEARCH_URL", "http://localhost:9200").split(
        ","
    )
    OVERRIDE_ELASTICSEARCH_URL = strtobool(
        os.environ.get("KARP5_ELASTICSEARCH_URL_OVERRIDE", "false")
    )
    TESTING = False
    DEBUG = False
    DATABASE_BASEURL = (
        f"mysql+pymysql://{ENV_DBPASS}/" + "{}?charset=utf8"
        if ENV_DBPASS
        else f"mysql+pymysql://{_CONFIG['DB']['DBPASS']}/" + "{}?charset=utf8"
    )
    STANDARDMODE = _CONFIG["SETUP"]["STANDARDMODE"]



"""Mode"""
from abc import ABCMeta, abstractmethod
import copy
from typing import Dict, List, Optional

import attr


def falsly_values_to_none(x):
    if not x:
        return None
    return x


@attr.s(auto_attribs=True)
class Mode:
    id: str = attr.ib()
    type: str = attr.ib()
    indexalias: str = attr.ib()
    suggestionalias: str = attr.ib()
    elastic_url: List[str] = attr.Factory(list)
    sql: Optional[str] = None
    secret_fields: List[str] = attr.Factory(list)
    src: Optional[str] = attr.ib(default=None, converter=falsly_values_to_none)
    sort_by: List[str] = attr.Factory(list)
    head_sort_field: List[str] = attr.Factory(list)
    autocomplete_field: List[str] = attr.Factory(list)
    minientry_fields: List[str] = attr.Factory(list)
    statistics_buckets: List[str] = attr.Factory(list)
    boosts: List[str] = attr.Factory(list)
    groups: List[str] = attr.Factory(list)
    is_index: bool = False
    filter_for_unauth_user: Optional[Dict] = None

    @classmethod
    def from_mapping(cls, id: str, mapping: Dict, defaults: Optional[Dict] = None):
        if defaults:
            kwargs = copy.deepcopy(defaults)
            kwargs.update(mapping)
        else:
            kwargs = dict(mapping)

        return cls(id=id, **kwargs)


class ModeRepository(metaclass=ABCMeta):
    @abstractmethod
    def mode_by_id(self, mode_id: str) -> Mode:
        raise NotImplementedError()

    @abstractmethod
    def mode_ids(self) -> List[str]:
        raise NotImplementedError()

    def modes_that_include_mode(self, mode_id: str) -> List[Mode]:
        modes = dict()
        modes[mode_id] = self.mode_by_id(mode_id)
        for m_id in self.mode_ids():
            if m_id == mode_id:
                continue
            mode = self.mode_by_id(m_id)
            if mode_id in mode.groups:
                modes[m_id] = mode

        return list(modes.values())



from pathlib import Path
from karp5.infrastructure.json_file_mode_repository import JsonFileModeRepository


def test_json_file_mode_repository():
    repo = JsonFileModeRepository(config_dir=Path("karp5/tests/data/config"))

    assert repo.mode_ids() == ["panacea", "panacea_links", "karp", "foo", "large_lex"]

    for mode_id in repo.mode_ids():
        mode = repo.mode_by_id(mode_id)

        assert mode.id == mode_id



import json
import time
import os

from urllib.request import urlopen
from itertools import zip_longest

# import elasticsearch_dsl as es_dsl
import pytest

from karp5.cli import upload_offline as upload


def get_es_indices():
    url = "http://127.0.0.1:9201/_cat/indices?format=json&h=health,index,docs.count"
    return json.loads(urlopen(url).read().decode("utf-8"))


def get_es_aliases():
    url = "http://127.0.0.1:9201/_cat/aliases?format=json&h=alias,index"
    return json.loads(urlopen(url).read().decode("utf-8"))


def get_es_search(alias):
    url = "http://127.0.0.1:9201/{}/_search".format(alias)
    return json.loads(urlopen(url).read().decode("utf-8"))


def mk_indexname(mode, suffix):
    return "{}_{}".format(mode, suffix)


def _test_index_exists(mode, suffix, n_docs):
    if isinstance(n_docs, int):
        n_docs = str(n_docs)
    es_status = get_es_indices()
    print(es_status)
    index = mk_indexname(mode, suffix)
    for index_status in es_status:
        if index_status["index"] == index:
            assert index_status["health"] == "yellow" or "green"
            assert index_status["docs.count"] == n_docs
            break
    else:
        assert False, "Didn't find index '{}'".format(index)


def _test_alias_contains_index(alias, index):
    es_status = get_es_aliases()
    print(es_status)
    for x in es_status:
        if x["alias"] == alias and x["index"] == index:
            break
    else:
        assert False, "Didn't find index '{}' in alias '{}'".format(index, alias)


def _test_n_hits_equals(mode, n_hits):
    es_hits = get_es_search(mode)
    print(es_hits)
    assert "hits" in es_hits
    assert "hits" in es_hits["hits"]
    assert "total" in es_hits["hits"]
    assert es_hits["hits"]["total"] == n_hits


def test_create_empty_index(cli_w_es):
    mode = "panacea"
    suffix = "test_upload_01"
    r = cli_w_es.create_empty_index(mode, suffix)

    assert r.exit_code == 0

    _test_index_exists(mode, suffix, "0")
    _test_n_hits_equals(mk_indexname(mode, suffix), 0)
    assert upload.index_exists(mode, mk_indexname(mode, suffix))


def test_lexicon_init(cli_w_es):
    mode = "foo"
    lexicon = "bar"
    suffix = "test_init_01"
    result = cli_w_es.lexicon_init(lexicon, suffix)

    assert result.exit_code == 0

    time.sleep(5)
    _test_index_exists(mode, suffix, 3)

    filename = os.path.join("karp5", "tests", "data", "data", "custom_path_bar.json")
    result = cli_w_es.lexicon_init(lexicon, suffix, data=filename)

    assert result.exit_code == 0

    time.sleep(5)
    _test_index_exists(mode, suffix, 5)


def test_create_reindex_alias(cli_w_panacea):
    mode = "panacea"
    suffix = "test_upload_02"
    n_hits = 6609
    time.sleep(5)
    _test_n_hits_equals(mode, n_hits)
    r = cli_w_panacea.reindex_alias(mode, suffix)

    assert r.exit_code == 0

    time.sleep(10)
    _test_index_exists(mode, suffix, str(n_hits))

    r = cli_w_panacea.publish_mode(mode, suffix)
    assert r.exit_code == 0
    time.sleep(10)

    _test_alias_contains_index(mode, mk_indexname(mode, suffix))
    _test_alias_contains_index("karp", mk_indexname(mode, suffix))
    _test_alias_contains_index("panacea_links", mk_indexname(mode, suffix))
    _test_n_hits_equals(mode, n_hits)
    _test_n_hits_equals("karp", n_hits)
    _test_n_hits_equals("panacea_links", n_hits)


def test_copy_mode(cli_w_panacea):
    source_mode = "panacea"
    target_mode = "panacea_links"
    target_suffix = "test_upload_03"
    n_hits = 6609
    _test_n_hits_equals(source_mode, n_hits)
    ok, errors = upload.copy_alias_to_new_index(source_mode, target_mode, target_suffix)

    assert ok
    assert errors is None

    time.sleep(5)
    _test_index_exists(target_mode, target_suffix, n_hits)

    r = cli_w_panacea.publish_mode(target_mode, target_suffix)
    assert r.exit_code == 0
    time.sleep(5)

    _test_alias_contains_index(target_mode, mk_indexname(target_mode, target_suffix))
    _test_n_hits_equals(source_mode, n_hits)
    _test_n_hits_equals(mk_indexname(target_mode, target_suffix), n_hits)
    _test_n_hits_equals(target_mode, 2 * n_hits)


def test_copy_mode_w_query(cli_w_panacea):
    source_mode = "panacea"
    target_mode = "panacea_links"
    target_suffix = "test_upload_04"
    source_n_hits = 6609
    target_n_hits = 1677
    query = {"query": {"match": {"pos_german": "Vb"}}}
    # query = es_dsl.Q('match', pos='Vb')
    _test_n_hits_equals(source_mode, source_n_hits)
    ok, errors = upload.copy_alias_to_new_index(
        source_mode, target_mode, target_suffix, query=query
    )

    assert ok
    assert errors is None

    time.sleep(5)
    _test_index_exists(target_mode, target_suffix, target_n_hits)

    r = cli_w_panacea.publish_mode(target_mode, target_suffix)
    assert r.exit_code == 0
    time.sleep(5)

    _test_alias_contains_index(target_mode, mk_indexname(target_mode, target_suffix))
    _test_n_hits_equals(source_mode, source_n_hits)
    _test_n_hits_equals(mk_indexname(target_mode, target_suffix), target_n_hits)
    _test_n_hits_equals(target_mode, source_n_hits + target_n_hits)


def mock_copy_alias(source_docs, lex, filter_func=None):
    if filter_func:
        source_docs = upload.apply_filter(source_docs, filter_func)

    def update_doc(doc):
        doc["_source"]["lexicon"] = lex
        return doc

    updated_docs = (update_doc(doc) for doc in source_docs)

    return updated_docs


def gen_data(n):
    for i in range(1, n):
        yield {"_source": {"id": i}, "_id": i + 4321}


def filter_even_gen(doc, _id):
    if doc["id"] % 2 == 0:
        yield doc
        c = doc.copy()
        c["link"] = c["id"]
        c["id"] = c["id"] + 100
        yield c


def filter_even_list(doc, _id):
    result = []
    if doc["id"] % 2 == 0:
        c = doc.copy()
        c["link"] = c["id"]
        c["id"] = c["id"] + 100
        result.append(c)
    return result


@pytest.mark.parametrize(
    "gen,lex,filter_func,expected",
    [
        (gen_data(10), "lex", None, gen_data(10)),
        (
            gen_data(10),
            "lex",
            filter_even_gen,
            [
                {"_source": {"id": 2}},
                {"_source": {"id": 102, "link": 2}},
                {"_source": {"id": 4}},
                {"_source": {"id": 104, "link": 4}},
                {"_source": {"id": 6}},
                {"_source": {"id": 106, "link": 6}},
                {"_source": {"id": 8}},
                {"_source": {"id": 108, "link": 8}},
            ],
        ),
        (
            gen_data(10),
            "lex",
            filter_even_list,
            [
                {"_source": {"id": 102, "link": 2}},
                {"_source": {"id": 104, "link": 4}},
                {"_source": {"id": 106, "link": 6}},
                {"_source": {"id": 108, "link": 8}},
            ],
        ),
    ],
)
def test_apply_filter(gen, lex, filter_func, expected):
    docs = mock_copy_alias(gen, lex, filter_func)

    for x, f in zip_longest(docs, expected):
        assert x is not None
        assert f is not None

        assert x["_source"]["id"] == f["_source"]["id"]
        assert x["_source"]["lexicon"] == lex


def test_recover(cli_w_panacea):
    alias = "panacea"
    alias_hits = get_es_search(alias)

    suffix = "test_recover"
    status_is_ok, indexname = upload.recover(alias, suffix, alias)

    assert status_is_ok
    time.sleep(3)
    recover_hits = get_es_search(indexname)

    assert "hits" in alias_hits
    assert "hits" in alias_hits["hits"]
    assert "total" in alias_hits["hits"]

    assert "hits" in recover_hits
    assert "hits" in recover_hits["hits"]
    assert "total" in recover_hits["hits"]

    assert alias_hits["hits"]["total"] == recover_hits["hits"]["total"]



import json
from pathlib import Path
from typing import Dict, List, Optional

from karp5.domain.model.mode import Mode, ModeRepository


class JsonFileModeRepository(ModeRepository):
    def __init__(self, config_dir: Path, filename: Optional[str] = None):
        self.config_dir = Path(config_dir)
        self.filename = "modes.json" if not filename else filename
        self.modes: Dict[str, Mode] = {}
        with open(self.config_dir / self.filename) as fp:
            modes_config = json.load(fp)

        self.defaults = modes_config.pop("default", None)
        for mode_id, mode_config in modes_config.items():
            self.modes[mode_id] = Mode.from_mapping(
                id=mode_id, mapping=mode_config, defaults=self.defaults
            )

    def mode_by_id(self, mode_id: str) -> Mode:
        return self.modes[mode_id]

    def mode_ids(self) -> List[str]:
        return list(self.modes.keys())



# -*- coding: utf-8 -*-
""" The backend, redirects url calls to the appropriate modules.
    Is also responsible for which ES node to query.
"""

import logging

from flask import jsonify, session

from karp5.server import checkdbhistory
import karp5.server.idgenerator as idgenerator
import karp5.server.searching as searching
import karp5.server.suggestions as suggestions
import karp5.server.update as update
from karp5.config import conf_mgr
import karp5


_logger = logging.getLogger("karp5")


def init(route):
    @route()
    def explain():
        """ Asking a query and show the elastic query """
        return searching.explain()

    @route()
    def test():
        """ Show the elastic query """
        return searching.test()

    @route()
    def query(page=0):
        """ Querying the database """
        return searching.query(page=page)

    @route()
    def querycount(page=0):
        """ Querying the database """
        return searching.querycount(page=page)

    @route()
    def minientry():
        """ Returns just some information about the results """
        return searching.minientry()

    @route()
    def statistics():
        """ Returns the counts and statistics """
        return searching.statistics()

    @route()
    def statlist():
        """ Returns the counts and statistics """
        return searching.statlist()

    @route()
    def random():
        return searching.random()

    @route()
    def autocomplete():
        return searching.autocomplete()

    # For seeing the posted data formated
    @route()
    def format():
        return searching.formatpost()

    # For getting a possible new identifier for an entry
    @route("<mode>")
    def suggestid(mode):
        return jsonify({"suggested_id": idgenerator.suggest_id(mode)})

    # For seeing the posted data formated
    @route("<lexicon>")
    def export(lexicon):
        return searching.export(lexicon)

    # For deleting a lexical entry from elastic and sql
    @route("<lexicon>/<_id>")
    def delete(lexicon, _id):
        return update.delete_entry(lexicon, _id, sql=True)

    # For updating a document
    @route("<lexicon>/<_id>", methods=["POST"])
    def mkupdate(lexicon, _id):
        return update.update_doc(lexicon, _id)

    # For adding a document
    @route("<lexicon>", methods=["POST"])
    def add(lexicon):
        return update.add_doc(lexicon)

    # For adding a document which already has an id (one that has been deleted)
    @route("<lexicon>/<_id>", methods=["POST"])
    def readd(lexicon, _id):
        return update.add_doc(lexicon, _id=_id)

    # For adding many document
    @route("<lexicon>", methods=["POST"])
    def addbulk(lexicon):
        return update.add_multi_doc(lexicon)

    @route("<lexicon>/<parentid>", methods=["POST"])
    def addchild(lexicon, parentid):
        return update.add_child(lexicon, parentid, suggestion=False)

    # For checking which resources a user may edit
    @route(methods=["GET"])
    def checkuser():
        return update.checkuser()

    # For retrieving update history of an entry
    @route("<lexicon>/<_id>")
    def checkhistory(lexicon, _id):
        return checkdbhistory.checkhistory(lexicon, _id)

    # For retrieving update history of a user
    @route()
    def checkuserhistory():
        return checkdbhistory.checkuserhistory()

    # For retrieving update history of one lexicon
    @route("<lexicon>")
    @route("<lexicon>/<date>")
    def checklexiconhistory(lexicon, date=""):
        try:
            return checkdbhistory.checklexiconhistory(lexicon, date=date)
        except Exception as e:
            raise e

    # For retrieving the lexicon order of a lexicon
    @route()
    def lexiconorder():
        return searching.lexiconorder()

    # For retrieving the difference between two versions
    @route("<lexicon>/<_id>/latest")
    @route("<lexicon>/<_id>/latest/<fromdate>")
    @route("<lexicon>/<_id>/<fromdate>/<todate>")
    def checkdifference(_id, lexicon, todate="", fromdate=""):
        return checkdbhistory.comparejson(lexicon, _id, todate=todate, fromdate=fromdate)

    # For submitting a suggestion
    @route(name="/suggestnew/<lexicon>", methods=["POST"])
    @route("<lexicon>/<_id>", methods=["POST"])
    def suggest(lexicon, _id=None):
        return suggestions.suggest(lexicon, _id)

    # For seeing suggestions
    @route()
    def checksuggestions():
        return suggestions.checksuggestions()

    # For accepting a suggestion
    @route("<lexicon>/<_id>", methods=["POST"])
    def acceptsuggestion(lexicon, _id):
        return suggestions.acceptsuggestion(lexicon, _id)

    # For accepting a suggestion
    @route("<lexicon>/<_id>", methods=["POST"])
    def acceptandmodify(lexicon, _id):
        return suggestions.acceptmodified(lexicon, _id)

    # For rejecting a suggestion
    @route("<lexicon>/<_id>", methods=["POST"])
    def rejectsuggestion(lexicon, _id):
        return suggestions.rejectsuggestion(lexicon, _id)

    # For seeing the status of a suggestion
    @route("<lexicon>/<_id>")
    def checksuggestion(lexicon, _id):
        return suggestions.checksuggestion(lexicon, _id)

    # For seeing entries that are the alphabetically close
    @route("<lexicon>")
    def getcontext(lexicon):
        return searching.get_context(lexicon)

    @route("<mode>")
    def modeinfo(mode):
        """ Show information about a mode """
        return searching.modeinfo(mode)

    @route("<lexicon>")
    def lexiconinfo(lexicon):
        """ Show information about a lexicon """
        return searching.lexiconinfo(lexicon)

    @route()
    def modes():
        jsonmodes = {}
        for mode, info in conf_mgr.modes.items():
            jsonmodes[mode] = info.get("groups", {})
        return jsonify(jsonmodes)

    @route()
    def groups():
        modes = {}
        for name, val in conf_mgr.lexicons.items():
            if name == "default":
                continue
            if val["mode"] in modes:
                modes[val["mode"]].append("%s (%s)" % (name, val["order"]))
            else:
                modes[val["mode"]] = ["%s (%s)" % (name, val["order"])]
        olist = ""
        for mode, kids in modes.items():
            olist += "<li>%s<ul>%s</ul></li>" % (
                mode,
                "\n".join("<li>%s</li>" % kid for kid in kids),
            )
        return "<ul> %s </ul>" % olist

    # ------------------- HTML FILES ------------------- #
    # Other ways of finding and sending files:
    # http://flask.pocoo.org/docs/0.10/api/#flask.Flask.open_resource
    # http://flask.pocoo.org/docs/0.10/api/#flask.Flask.send_static_file
    @route()
    def order():
        orderlist = []
        for name, val in conf_mgr.lexicons.items():
            orderlist.append((val["order"], "%s (%s)" % (name, val[0])))
        olist = "\n".join("<li>%d: %s</li>" % on for on in sorted(orderlist))
        return "<ul> %s </ul>" % olist

    @route(name="/")
    @route(name="/index")
    def helppage():
        """Render API documentation."""
        _logger.debug("ok")
        import markdown

        _logger.debug("index page")

        KARP_API_URL = conf_mgr.app_config.BACKEND_URL
        KARP_VERSION = "5"
        STYLES_CSS = "static/api.css"
        _logger.debug("abs path: %s", conf_mgr.app_config.ABSOLUTE_PATH)

        # doc_dir = os.path.join(conf_mgr.app_config.ABSOLUTE_PATH, 'src', 'html')
        # doc_file = 'api.md'
        #
        # with app.open_resource(os.path.join(doc_dir, doc_file)) as doc:
        #     md_text = doc.read()
        #     _logger.debug("md_text: %s", type(md_text))
        #     md_text = md_text.decode("UTF-8")
        #     _logger.debug("md_text: %s", type(md_text))

        md_text = karp5.get_pkg_resource("html/api.md")
        _logger.debug("md_text: %s", type(md_text))
        # Replace placeholders
        md_text = md_text.replace("[SBURL]", KARP_API_URL)
        md_text = md_text.replace("[SBVERSION]", conf_mgr.version)
        # md_text = md_text.replace("[URL]", request.base_url)
        # md_text = md_text.replace("[VERSION]", KARP_VERSION)

        # Convert Markdown to HTML
        md = markdown.Markdown(
            extensions=[
                "markdown.extensions.toc",
                "markdown.extensions.smarty",
                "markdown.extensions.def_list",
                "markdown.extensions.tables",
            ]
        )
        md_html = md.convert(md_text)
        md_html = md_html.replace("<pre><code>", '<pre><code class="json">')

        html = [
            """<!doctype html>
            <html>
              <head>
                <meta charset="utf-8">
                <title>Karp API v%s</title>
                <link href="//cdnjs.cloudflare.com/ajax/libs/highlight.js/9.12.0/styles/monokai-sublime.min.css"
                  rel="stylesheet">
                <script src="//cdnjs.cloudflare.com/ajax/libs/highlight.js/9.12.0/highlight.min.js"></script>
                <script>hljs.initHighlightingOnLoad();</script>
                <link href="https://fonts.googleapis.com/css?family=Roboto" rel="stylesheet">
                <link href="https://fonts.googleapis.com/css?family=Roboto+Slab" rel="stylesheet">
                <link href="%s" rel="stylesheet">
              </head>
              <body>
                <div class="toc-wrapper">
                  <div class="header">
                    <img src="static/karp.png"><br><br>
                    Karp API <span>v%s</span>
                  </div>
                  %s
                </div>
               <div class="content">
                """
            % (KARP_VERSION, STYLES_CSS, KARP_VERSION, md.toc),
            md_html,
            "</div></body></html>",
        ]

        return "\n".join(html)

    @route(name="/logout")
    def logout():
        # remove the username from the session if it's there
        session.pop("username", None)
        session.pop("lexicon_list", None)
        return jsonify({"logged_out": True})



from elasticsearch import exceptions as esExceptions
from karp5 import errors
from karp5.dbhandler.dbhandler import dbselect
from flask import request, jsonify
from json import loads
import logging
from karp5.context import auth
import karp5.server.helper.helpers as helpers
from karp5.config import mgr as conf_mgr
import karp5.server.update as update


_logger = logging.getLogger("karp5")


def suggest(lexicon, _id):
    return update.add_doc(lexicon, _id=_id, suggestion=True)


def checksuggestions():
    _, permitted = auth.validate_user()
    settings = {"allowed": permitted}
    helpers.get_querysettings(settings)
    size = settings.get("size", 50)
    lexicons = settings.get("resource", [])
    status = settings.get("status", ["waiting", "rejected", "accepted"])
    _logger.debug("checksuggestions in %s" % lexicons)
    if not lexicons:
        return jsonify({"updates": []})
    updates = []
    for lexicon in lexicons:
        # add updates from lexicons that are kept in sql
        if conf_mgr.get_lexicon_sql(lexicon):
            updates.extend(dbselect(lexicon, suggestion=True, status=status, max_hits=size))

    return jsonify({"updates": updates})


def acceptsuggestion(lexicon, _id):
    try:
        ans = savesuggestion(lexicon, _id)
        _logger.debug("saved sugg")
        return jsonify(ans)
    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        update.handle_update_error(e, {"id": _id}, helpers.get_user(), "accept")
        raise errors.KarpElasticSearchError(
            "Error during update. Document not saved.", debug_msg=str(e)
        )
    except Exception as e:
        update.handle_update_error(e, {"id": _id}, helpers.get_user(), "accept")
        raise errors.KarpGeneralError(str(e))


def acceptmodified(lexicon, _id):
    try:
        request.get_data()
        data = loads(request.data)
        modified_doc = data
        ans = savesuggestion(lexicon, _id, status="accepted_modified", source=modified_doc)
        return jsonify(ans)
    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        _logger.exception(e)
        update.handle_update_error(
            e, {"id": _id, "data": data}, helpers.get_user(), "accept modified"
        )
        raise errors.KarpElasticSearchError(
            "Error during update. Document not saved.", debug_msg=str(e)
        )
    except Exception as e:
        _logger.exception(e)
        update.handle_update_error(
            e, {"id": _id, "data": data}, helpers.get_user(), "accept modified"
        )
        raise errors.KarpGeneralError(str(e))


def savesuggestion(lexicon, _id, status="accepted", source=""):
    sugg_index, typ = conf_mgr.get_lexicon_suggindex(lexicon)
    es = conf_mgr.elastic(lexicon=lexicon)
    suggestion = es.get(index=sugg_index, doc_type=typ, id=_id)
    _, permitted = auth.validate_user()
    set_lexicon = suggestion["_source"]["lexiconName"]
    helpers.check_lexiconName(lexicon, set_lexicon, "rejectsuggestion", _id)
    if lexicon not in permitted:
        raise errors.KarpAuthenticationError("You are not allowed to update lexicon %s" % lexicon)

    origin = dbselect(lexicon, suggestion=True, _id=_id, max_hits=1)[0]
    origid = origin["origid"]
    request.get_data()
    data = loads(request.data)
    message = data.get("message")
    suggestion["message"] = message
    suggestion["version"] = origin["version"]
    if not source:
        source = suggestion
    # the user log in is checked in add_doc
    # add_doc raises exception if ES
    if origid:
        # update in ES
        ans = update.update_doc(lexicon, origid, data=source, live=False)
    else:
        # add to ES
        ans = update.add_doc(lexicon, live=False, data=source)
        origid = ans.get("_id")
    # mark as accepted
    ok, err = update.modify_db(_id, lexicon, message, status, origid=origid)
    # delete from suggestion index
    suggans = update.delete_entry(lexicon, _id, sql=False, live=False, suggestion=True)
    ans["sugg_db_loaded"] = ok
    ans["sugg_es_ans"] = suggans
    if not ok:
        _logger.debug(err)
    update.send_notification(origin["user"], message, _id, status)
    return ans


def rejectsuggestion(lexicon, _id):
    try:
        origin = dbselect(lexicon, suggestion=True, _id=_id, max_hits=1)[0]
    except Exception as e:
        # if error occurs here, the suggestion is not in sql
        raise errors.KarpDbError("Rejection not found", "Rejection not found: %s" % str(e))
    _, permitted = auth.validate_user()
    set_lexicon = origin["doc"]["lexiconName"]
    helpers.check_lexiconName(lexicon, set_lexicon, "rejectsuggestion", _id)
    if lexicon not in permitted:
        raise errors.KarpAuthenticationError("You are not allowed to update lexicon %s" % lexicon)
    try:
        origin = dbselect(lexicon, suggestion=True, _id=_id, max_hits=1)[0]
        # delete from suggestion index
        # the user log in is checked in delete_entry
        # delete_entry raises exception if ES fails
        sugg_index, typ = conf_mgr.get_lexicon_suggindex(lexicon)

        ans = update.delete_entry(lexicon, _id, sql=False, live=False, suggestion=True)
        request.get_data()
        data = loads(request.data)
        message = data.get("message")
        # mark as rejected
        ok, err = update.modify_db(_id, lexicon, message, "rejected")

        ans["sugg_db_loaded"] = ok
        if not ok:
            _logger.debug(err)
        update.send_notification(origin["user"], message, _id, "rejected")
        return jsonify(ans)
    except (esExceptions.RequestError, esExceptions.TransportError) as e:
        update.handle_update_error(e, {"id": _id}, helpers.get_user(), "reject")
        raise errors.KarpElasticSearchError(
            "Error during update. Document not saved.", debug_msg=str(e)
        )
    except Exception as e:
        update.handle_update_error(e, {"id": _id}, helpers.get_user(), "reject")
        raise errors.KarpGeneralError(str(e))


def checksuggestion(lexicon, _id):
    # TODO add exception handling
    try:
        return jsonify({"updates": dbselect(lexicon, suggestion=True, _id=_id, max_hits=1)})
    except Exception as e:
        raise errors.KarpGeneralError(str(e))
